/* This file is intended to be part of GTK - The GIMP Toolkit
 * Copyright (C) 1999 Eric S. Tiedemann (est@hyperreal.org)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "gtklevelhold.h"

static GtkRgbClass *parent_class;

static void
gtk_level_hold_size_allocate (GtkWidget     *widget,
				GtkAllocation *allocation)
{
  GtkLevelHold *lh;
  GtkRgb *rgb;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_LEVEL_HOLD (widget));
  g_return_if_fail (allocation != NULL);

  /* chain up first to update rgb->{h,w} */
  if (GTK_WIDGET_CLASS(parent_class)->size_allocate)
    GTK_WIDGET_CLASS(parent_class)->size_allocate(widget, allocation);

  lh = GTK_LEVEL_HOLD(widget);

  rgb = GTK_RGB(widget);

  g_free(lh->data);
  if (lh->options & GTK_LEVEL_HOLD_HORIZONTAL)
    lh->data = g_new0(int, rgb->w);
  else
    lh->data = g_new0(int, rgb->h);
}

static void
gtk_level_hold_destroy(GtkObject *object)
{
  GtkLevelHold *lh;

  g_return_if_fail(object != NULL);
  g_return_if_fail(GTK_IS_LEVEL_HOLD(object));

  lh = GTK_LEVEL_HOLD(object);

  g_free(lh->data);
  lh->data = 0;

  /*fprintf(stderr, "destroying levelhold\n");*/
  if (GTK_OBJECT_CLASS(parent_class)->destroy)
    GTK_OBJECT_CLASS(parent_class)->destroy(object);
}

static void
gtk_level_hold_class_init(GtkLevelHoldClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  /* we'll need this stuff */
  gdk_rgb_init();

  parent_class = gtk_type_class(gtk_rgb_get_type());

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;

  object_class->destroy = gtk_level_hold_destroy;
  widget_class->size_allocate = gtk_level_hold_size_allocate;
}

static void
gtk_level_hold_init(GtkLevelHold *lh)
{
}

guint
gtk_level_hold_get_type()
{
  static guint mytype = 0;

  if (!mytype)
    {
      GtkTypeInfo lh_info =
      {
	"GtkLevelHold",
	sizeof(GtkLevelHold),
	sizeof(GtkLevelHoldClass),
	(GtkClassInitFunc) gtk_level_hold_class_init,
	(GtkObjectInitFunc) gtk_level_hold_init,
	(GtkArgSetFunc) NULL,
	(GtkArgGetFunc) NULL
      };

      mytype = gtk_type_unique (gtk_rgb_get_type(), &lh_info);
    }

  return mytype;
}

GtkWidget*
gtk_level_hold_new(unsigned h, unsigned w, int options, unsigned steps, 
		   float hard_hold)
{
  GtkLevelHold *lh = gtk_type_new (gtk_level_hold_get_type ());
  GtkRgb *rgb = GTK_RGB(lh);

  rgb->h = h;
  rgb->w = w;
  rgb->can_resizep = !(options & GTK_LEVEL_HOLD_FIXED);

  /* request a pleasant default size */
  GTK_WIDGET(lh)->requisition.width = rgb->w;
  GTK_WIDGET(lh)->requisition.height = rgb->h;

  lh->hard_hold = CLAMP(hard_hold, 0.0, 1.0);
  lh->options = options;
  lh->steps = steps;

  return GTK_WIDGET (lh);
}

static void
gtk_level_hold_vertical_paint(GtkLevelHold *lh, float levf)
{
  int i;
  guchar *p, *sent;
  int lev;
  GtkRgb *rgb = GTK_RGB(lh);

  /* fade */
  lev = (int) floor(((rgb->h - 1) * CLAMP(levf, 0.0, 1.0)));
  
  for (i = 0; i < rgb->h; i++)
    if (lh->data[i] != 0)
      lh->data[i]--;
  
  /* set lev rows to max data level */
  for (i = 0; i <= lev; i++)
    lh->data[i] = lh->steps - 1;

  for (i = 0; i < rgb->h; i++)
    {
      int r, g, b;

      /* calculate hue */
      r = ((i < (rgb->h/2)) ? ((i/(((float) (rgb->h/2)))) * 255) : (255));
      g = ((i < (rgb->h/2)) 
	   ? (255) 
	   : (((rgb->h - i)/((float) (rgb->h/2))) * 255));
      b = 0;

      if (lh->data[i] == 0 || i < lh->hard_hold * rgb->h)
	/* scale to data level if not above hard-hold point */
	{
	  if (lh->options & GTK_LEVEL_HOLD_FADE)
	    {
	      r = (r * lh->data[i]) / lh->steps;
	      g = (g * lh->data[i]) / lh->steps;
	      b = (b * lh->data[i]) / lh->steps;
	    }
	  else
	    /* immediately blacken if we're not fading */
	    {
	      if (lh->data[i] != lh->steps - 1)
		r = g = b = 0;
	    }
	}

      p = rgb->img + ((rgb->h - 1) - i) * 3 * rgb->w;
      sent = p + 3 * rgb->w;

      while(p < sent)
	{
	  *p++ = r;
	  *p++ = g;
	  *p++ = b;
	}
    }
}

static void
gtk_level_hold_horizontal_paint(GtkLevelHold *lh, float levf)
{
  int i, j;
  guchar *p;
  int lev;
  GtkRgb *rgb = GTK_RGB(lh);

  /* fade */
  lev = (int) floor(((rgb->w - 1) * CLAMP(levf, 0.0, 1.0)));

  for (i = 0; i < rgb->w; i++)
    if (lh->data[i] != 0)
      lh->data[i]--;

  /* set lev rows to max data level */
  for (i = 0; i <= lev; i++)
    lh->data[i] = lh->steps - 1;

  for (i = 0; i < rgb->w; i++)
    {
      int r, g, b;

      /* calculate hue */
      r = ((i < (rgb->w/2)) ? ((i/(((float) (rgb->w/2)))) * 255) : (255));
      g = ((i < (rgb->w/2)) 
	   ? (255) 
	   : (((rgb->w - i)/((float) (rgb->w/2))) * 255));
      b = 0;

      if (lh->data[i] == 0 || i < lh->hard_hold * rgb->w)
	/* scale to data level if not above hard-hold point */
	{
	  if (lh->options & GTK_LEVEL_HOLD_FADE)
	    {
	      r = (r * lh->data[i]) / lh->steps;
	      g = (g * lh->data[i]) / lh->steps;
	      b = (b * lh->data[i]) / lh->steps;
	    }
	  else
	    /* immediately blacken if we're not fading */
	    {
	      if (lh->data[i] != lh->steps - 1)
		r = g = b = 0;
	    }
	}

      p = rgb->img + 3 * i;
      for(j = 0; j < rgb->h; j++, p += 3 * rgb->w)
	{
	  p[0] = r;
	  p[1] = g;
	  p[2] = b;
	}
    }
}

void
gtk_level_hold_set(GtkLevelHold *lh, gfloat levf)
{
  if (lh->options & GTK_LEVEL_HOLD_HORIZONTAL)
    gtk_level_hold_horizontal_paint(lh, levf);
  else
    gtk_level_hold_vertical_paint(lh, levf);

  gtk_widget_draw(GTK_WIDGET(lh), 0);
}

/*#define TEST*/
#ifdef TEST

gint
delete_event(GtkWidget *w, GdkEvent *e, gpointer o)
{
  gtk_main_quit();
}

float phase;
float phase1;

gint
bar(gpointer o)
{
  GtkLevelHold *lh = GTK_LEVEL_HOLD(o);

  gtk_level_hold_set(lh, (sin(phase) * cos(phase1) + 1) / 2);
  phase += .03;
  phase1 += .07;
}

int
main(int argc, char **argv)
{
  GtkWidget *w, *lh;

  gtk_init(&argc, &argv);

  w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  
  gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, TRUE);

  gtk_signal_connect(GTK_OBJECT(w), "delete_event", 
		     GTK_SIGNAL_FUNC(delete_event), (gpointer) 0);

  lh = gtk_level_hold_new(200, 10, 
			  GTK_LEVEL_HOLD_FADE,
			  100,
			  .9);

  //lh = gtk_level_hold_new(200, 10, 
  //		  GTK_LEVEL_HOLD_FADE | GTK_LEVEL_HOLD_HORIZONTAL,
  //		  100,
  //		  .9);

  gtk_container_add(GTK_CONTAINER(w), GTK_WIDGET(lh));

  gtk_widget_show_all(w);

  gtk_timeout_add(10, bar, lh);

  gtk_main();

  return 0;
}
#endif
